/*================================================================*/
/*
* Parallel Recursive Spectral Bisection -
* Based on the sequential version provided by Horst Simon
* -ravi
*/
/*================================================================*/

#include <stdio.h>
#include "hum.h"
#include "parti.h"

/*
#include <util.h>
#define malloc(size) NEWN(char, size)
#define free(ptr) DISPOSE(ptr)
*/

/*================================================================*/
/* called by genord() 
 * finds fiedler value and the coressponding vector 
 */
/*================================================================*/

void
lanczo(sched,  xadj_ind, xadj,  adj_loc, neqns, n_off_proc, 
       diag, eival, evec, u,  teqns)
SCHED *sched;
int *xadj_ind, *xadj, *adj_loc;
int neqns, teqns;
int n_off_proc;
double *eival;
double *diag, *evec, *u;
{
   double *alf, *r;
   int i,  j, jmax,  iflag;
   double res, alfa, beto, be2o, eps, maceps, dif, du;
   double  xc, xl, bd, xa, w, xb, da, db, sumi, tol;
   int lanstep; 
   double dl, eig, xu, eval;
   int ic, conv, converged, ierror;
   double nrm2();
   double dot();
   double root();
   void scal(), copy(), axpy(), swap(), pivot(), givens();
   
#ifdef DEBUG2
    if(PARTI_mynode == 0) printf("entered lanczos \n"); PARTI_gsync();
#endif

   iflag     = 0;
   maceps    = 1.0e-12;
   eps       = 0.0001;
   conv      = 0;
   converged = 0;

   bet[0]  = 0.0;
   betc[0] = 0.0;

  
  alf = (double *) malloc(sizeof(double)*MAXLAN);
  r   = (double *) malloc(sizeof(double)*(neqns+n_off_proc));

  for(i=1; i<=neqns; i++) {
    r[i-1] = 0.0;
    u[i-1] = 0.498 * (double)(-teqns+2*(xadj_ind[i-1]+1));
  }
  for(i=0; i<n_off_proc; i++) {
    r[i+neqns] = 0.0;
  }
  ortho(neqns, u, teqns);

  res = nrm2(neqns, u);
  res = 1.0e0/res;

#ifdef DEBUG2
  if (PARTI_mynode == 0) printf(" --> RES[%d] %f  \n", PARTI_mynode, res); PARTI_gsync();
#endif

  scal(neqns, res, u);
  copy(neqns, u, evec);
  
  sumi = 0.0;
  lanstep = 0;

  for(j=0; j<MAXLAN; j++) {
/*
    if (PARTI_mynode == 0) 
    printf("************* LOOP J %d **********************\n", j);
*/
    scal(neqns, -bet[j], r);
    opk(sched, xadj, adj_loc,diag,u,r, neqns, n_off_proc);
    alf[j] = dot(neqns, u, r);
    axpy(neqns, -alf[j], u, r);
    
    ortho(neqns, r, teqns);

    bet[j+1] = nrm2(neqns, r);

    if (bet[j+1] <= maceps ) {
       printf("small beta encountered bet[%d]: %f \n", j, bet[j+1]);
       iflag = 1;
       bet[j+1] = 0.00001;
    }
    scal (neqns, 1.0/bet[j+1], r);
    swap(neqns, u, r);

    alfa = alf[j];
    beto = bet[j];
    be2o = beto*beto;
    betc[j] = be2o;
    tol = 0.5*eps*fabs(eig);

    if(j == 1) {
       dif = (alf[0] - alfa) *0.5;
       xc = alfa + dif + sqrt((dif*dif)+be2o);
       xu = xc + eps * fabs(xc);
       du = alfa - (be2o/(alf[0]-xu)) - xu;
       xl = 2.0*xc - xu;
       eig = xc;
    }
    else if( j > 1) {
       du = alfa - xu -be2o / du;
       ic = 0;
       if(du > 0.0)  {
          xl = xu;
          xa = xl + bd;
          pivot(alf, &betc[1], j, xa, &da);
          ic++;
          eig = xc;
          xb  = xl + 0.5*(xa - xl);
          xu  = beto + MAX(xl, alfa);
          
          for(conv=0,i=0; i<MAXITER; i++) {
             pivot(alf, &betc[1], j, xb, &db);
             ic++;
             if (db > 0.0) {
               xl = xb;
             }
             else {
               xu = xb;
               du = db;
             }
             w = xb - xa;
             if (w <= 0.0) {
                xc = xb;
             }
             else {
                xc = xb + root(-w-(db*(xb-eig)-da*(xa-eig))/w,db*(xb-eig));

             }
             if (xc < (xl+0.5*tol) || xc > (xu - 0.5*tol)) {
                xc = xl+0.5*(xu-xl);
             }
             xa = xb; da = db; xb = xc;
             if ((xu-xl) <= eps) {
                conv = 1;
                 break;
             }
          }
       }
       else conv = 1;
    } 
   
       if (j > 1 && conv != 1) {
         ierror = 2;
       } 
     if ( j != 0) {
       if (du == 0.0) du = eps;


       givens(alf, bet, j, xc, z);
       res = fabs(z[j])*bet[j+1];
       bd = res;
       sumi = sumi+(double) (ic);

       jmax = j;
       lanstep = j;
       if (res <= eps) {
          converged = 1;
          break;
       }
     }
  }

  if(converged != 1) {
     ierror = 1;
  }
  
/*
    if (PARTI_mynode == 0) 
    printf("\n************* OUT OF LANCZ LOOP J %d **********************\n", j);
*/
  eig = xc;
  sumi = sumi/(double)jmax;
  copy(neqns, evec, u);
  for(i=0; i<neqns; i++) {
    evec[i] = z[0]*u[i]; 
    r[i]    = 0.0;
  }
  eval = alf[0];
  bet[0] = 0.0;
  betc[0] = 0.0;

  for(j=0; j<jmax-1; j++) {
    scal(neqns, -bet[j], r);
/*
    if (PARTI_mynode == 0) printf("j = %d\n", j);
*/
    opk(sched, xadj, adj_loc,diag,u,r, neqns, n_off_proc);
    alf[j] = dot(neqns, u, r);
    axpy(neqns, -alf[j], u, r);

    ortho(neqns, r, teqns);

    bet[j+1] = nrm2(neqns, r);

    scal (neqns, 1.0/bet[j+1], r);
    swap(neqns, u, r);
    axpy(neqns, z[j+1], u, evec);
  }
 
   *eival = eval;

   free(alf); free(r);
   return;
}
/*================================================================*/
